import functools

import numpy as np
from matplotlib import pyplot as plt
from shapely import geometry as shg
from sortedcontainers import SortedDict

from lgblkb_tools import logger
from lgblkb_tools.geometry.utils.geom_utils import GenericGeometry, line_xy, perpendicular, normalize, min_dist
from .point_utils import ThePoint
from .utils.plot_utils import plot_patches


class TheLine(GenericGeometry):

    def __init__(self, coordinates=None, width=2.0, show='', line=None, text=None, **data):
        super(TheLine, self).__init__(**data)
        if isinstance(coordinates, shg.LineString):
            line = coordinates
        elif isinstance(coordinates, TheLine):
            line = coordinates.line
        elif isinstance(coordinates, np.ndarray):
            assert coordinates.shape[0] > 1
        elif isinstance(coordinates, list):
            assert len(coordinates) > 1
        else:
            raise NotImplementedError(f'type = {type(coordinates)}, value={coordinates}')
        try:
            self.line = line or shg.LineString(coordinates=coordinates)
        except AssertionError:
            logger.debug('coordinates:\n%s', coordinates)
            raise
        # self.line=get_rounded(self.line,decimals=round_decimals)
        # self._generate_id()
        self.width = width
        self._coverage = None
        self._xy = None
        self._vector = None
        self._unit_vector = None
        self._slope = None
        self._midpoint = None
        self._perp_unit_vector = None
        self._cover_line = None
        self._show = show
        self.text = text
        self.intermediate_points = SortedDict()
        # if show: self.plot(show,**plot_kwargs)
        pass

    @property
    def geometry(self):
        return self.line

    def get_all_points(self):
        # return [self[0],*self.intermediate_points,self[1]]
        return [self[0], *self.intermediate_points, self[1]]
        pass

    @property
    def coverage(self):
        if self._coverage is None:
            self._coverage = self.line.buffer(self.width, cap_style=shg.CAP_STYLE.round)
        return self._coverage

    def plot(self, showcode='', show_self=1, show_normal=0, show_coverage=0, show_text=None, text='', **kwargs):
        get_showcode = lambda: [int(x) if x is not None else None for x in showcode]
        if showcode:
            if len(showcode) == 1:
                show_self = int(showcode)
            elif len(showcode) == 2:
                show_self, show_normal = get_showcode()
            elif len(showcode) == 3:
                show_self, show_normal, show_coverage = get_showcode()
            elif len(showcode) == 4:
                show_self, show_normal, show_coverage, show_text = get_showcode()
        if show_self:
            plt.plot(*self.line.xy, **dict(dict(), **kwargs))
        if show_normal:
            norm_line = self.get_normal_line(as_the_line=False)
            plt.plot(norm_line[:, 0], norm_line[:, 1], c='y')
        if show_coverage:
            plot_patches([self.coverage], c='green', alpha=0.1)
        if (text or self.text) and show_text is None:
            # p1=plt.gca().transData.transform_point(self.xy[0])
            # p2=plt.gca().transData.transform_point(self.xy[1])
            # dy=abs(p2[1]-p1[1])
            # dx=abs(p2[0]-p1[0])
            # rotn=np.degrees(np.arctan2(dy,dx))
            # trans_angle=plt.gca().transData.transform_angles(
            # 	np.array((np.arctan(self.slope),)),self.midpoint.reshape((1,2)),radians=True)[0]
            # a=np.array(self.line.interpolate(0.5,normalized=True).xy)
            # a=a+(self.perp*3).reshape(a.shape)
            # print(a)
            # print((np.abs(self.perp)*5).reshape(a.shape))
            # plt.annotate(text or self.text,a,rotation=trans_angle*180/np.pi,rotation_mode='anchor')
            plt.text(*ThePoint(self.line.centroid), text or self.text)
        return self

    def get_normal_line(self, as_the_line=True, width=None):
        normal_vector = self.perp * (width or self.width)
        normal_line = np.array([self.midpoint - normal_vector, self.midpoint + normal_vector])
        return TheLine(normal_line, width=self.width) if as_the_line else normal_line

    def get_cover_line(self, cover_line_length):
        return self.get_normal_line().get_normal_line(width=cover_line_length)

    def get_field_lines(self, field_polygon, as_the_line=True):
        field_polygon = field_polygon.as_valid(field_polygon.geometry)
        some_line = self.get_cover_line(field_polygon.cover_line_length).line
        try:
            # field_polygon.plot()
            # plt.show()

            lines = field_polygon.polygon.intersection(some_line)
        except Exception as exc:
            logger.exception(str(exc))
            logger.debug('some_line:\n%s', some_line)
            logger.debug('field_polygon.polygon:\n%s', field_polygon.polygon)

            field_polygon.plot(lw=5, c='r')
            self.get_cover_line(field_polygon.cover_line_length).plot(lw=5, c='k')
            plt.show()
            # field_polygon:ThePoly=field_polygon
            # lines=field_polygon.as_valid(field_polygon).polygon.buffer(1).buffer(-1).intersection(self.get_cover_line(field_polygon.cover_line_length).line)
            raise exc

        if as_the_line:
            liner = functools.partial(TheLine, width=self.width)
        else:
            liner = lambda x: x

        if lines.is_empty:
            return []
        elif isinstance(lines, shg.MultiLineString):
            return [liner(x) for x in lines]
        elif isinstance(lines, shg.LineString):
            try:
                return [liner(lines)]
            except:
                logger.debug('lines: %s', lines)
                raise
        else:
            raise NotImplementedError(f"type(lines) = {type(lines)}, lines={lines}")

    def reversed(self):
        return TheLine([self[1], self[0]])

    @property
    def vector(self):
        if self._vector is None: self._vector = self[1] - self[0]
        return self._vector

    @property
    def unit_vector(self):
        if self._unit_vector is None: self._unit_vector = self.vector.xy / np.linalg.norm(self.xy)
        return self._unit_vector

    @property
    def xy(self):
        if self._xy is None:
            self._xy = line_xy(self.line)
        return self._xy

    @property
    def slope(self):
        if self._slope is None:
            self._slope = 1 / np.divide(*(self.xy[1, :] - self.xy[0, :]))
        return self._slope

    @property
    def perp(self):
        if self._perp_unit_vector is None:
            self._perp_unit_vector = perpendicular(normalize(self.xy[1, :] - self.xy[0, :]))
        return self._perp_unit_vector

    @property
    def midpoint(self):
        if self._midpoint is None:
            self._midpoint = np.sum(self.xy, 0) / 2
        return self._midpoint

    @property
    def line_points(self):
        return [*self[:], *self.intermediate_points]

    def get_along_normal(self, distance, point_on_line=None):
        point_on_line = self.midpoint if point_on_line is None else point_on_line
        return point_on_line + self.perp * distance

    def offset_by(self, distance, count=1):
        offset_lines = list()
        for i in range(count):
            points = self.get_along_normal(distance=distance * (1 + i), point_on_line=self.xy)
            line = TheLine(coordinates=points, width=self.width)
            offset_lines.append(line)
        return offset_lines

    def __eq__(self, other):
        if np.linalg.norm(self.xy - other.xy) < min_dist or np.linalg.norm(self.xy - other.xy[::-1]) < min_dist:
            return True
        else:
            return False

    def touches(self, xy):
        return self.line.touches(shg.Point(xy))

    def __len__(self):
        return len(self.xy)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [self[ii] for ii in range(*item.indices(len(self)))]
        # logger.debug('self.xy: %s',self.xy)
        # logger.debug('item: %s',item)
        # logger.debug('self.xy[item]: %s',self.xy[item])
        return ThePoint(self.xy[item])

    def __iter__(self):
        return iter([self[i] for i in range(2)])

    def extend_from(self, xy, length, show=False, **plot_kwargs):
        # ThePoint([0,0]).plot('Origin')
        # target_point=ThePoint(crop_field.xy[0]).plot()
        # target_line=TheLine([[0,50],[100,50]]).plot(text='target_line')
        # target_line=TheLine([[0,50],[-100,50]]).plot(text='target_line')
        # target_line=TheLine([[0,50],[0,100]]).plot(text='target_line')
        # target_line=TheLine([[0,50],[0,-100]]).plot(text='target_line')
        # target_point=ThePoint(target_line[0]).plot('Target')
        # target_point=ThePoint(self[0])  #.plot('Target')
        xy_point = ThePoint(xy)
        # if xy_point.point.distance(self[0].point)<min_dist:
        if xy_point.almost_touches(self[0]):  # .point.distance(self[0].point)<min_dist:
            target_point = self[0]
            line_vector = self[1] - self[0]
            _power = 0
        # elif xy_point.point.distance(self[1].point)<min_dist:
        elif xy_point.almost_touches(self[1]):  # point.distance(self[1].point)<min_dist:
            target_point = self[1]
            line_vector = self[0] - self[1]
            _power = 1
        else:
            message = 'xy_point does not correspond to any of the line vertices.'
            logger.error(message)
            # xy_point.plot('xy_point')
            # self.plot(text='The line')
            # plt.show()
            raise ValueError(message, dict(xy_point=str(xy_point), the_line=str(self)))

        # logger.debug('target_line.slope: %s',target_line.slope)
        # logger.debug('self.slope: %s',self.slope)
        # line_vector.plot('line_vector')
        # target_point.plot('Target')
        # self[0].plot('Start')
        # self[1].plot('End')
        # plt.show()
        if np.isposinf(self.slope) or np.isneginf(self.slope):
            # vector_point=ThePoint([0,(-1)**(np.isneginf(self.slope))*length])*(-1)**(self.vector.y>0)
            vector_point = ThePoint([0, (-1) ** _power * length])  # *(-1)**(self.vector.y>0)
        # vector_point.plot('Vector point')
        else:
            vector_point = ThePoint([length, self.slope * length]).unit * length  # .plot('Vector point')
        # line_vector.plot(text='line_vector')
        new_point = (target_point + ((-1) ** (line_vector.x > 0)) * vector_point)  # .plot('New point')
        extension_line = TheLine([target_point.xy, new_point.xy])  # .plot(text='extension_line')
        if show: extension_line.plot(**plot_kwargs)
        # plt.show()
        return extension_line

    def __contains__(self, item):
        # logger.debug('item: %s',item)
        if not isinstance(item, ThePoint): item = ThePoint(item)
        for point in self[:]:
            # logger.debug('type(point): %s',type(point))
            # logger.debug('point: %s',point)
            if point == item: return True
        return False

    def __hash__(self):
        return hash(self.geometry)

    @property
    def angle(self):
        return np.arctan2(self.vector.y, self.vector.x) * 180 / np.pi

# def __iter__(self):
# 	return iter(self.xy)



import itertools as it

import matplotlib.pyplot as plt
import numpy as np
from box import Box


class Plotter(object):
    def __init__(self, *objs, axis_off=True, as_images=True, **plot_kwargs):
        self.data = list()
        self.axis_off = axis_off
        if objs:
            if as_images:
                self.add_images(*objs).plot(**plot_kwargs).show()
            else:
                self.add_plots(*objs).plot(**plot_kwargs).show()
        pass
    
    def _add_obj(self, obj, ax_callback, plot_kwargs, **kwargs):
        self.data.append(Box(obj=obj, plot_kwargs=plot_kwargs, ax_callback=ax_callback, **kwargs))
        return self
    
    # region Front:
    def add_image(self, obj=None, ax_callback=None, **plot_kwargs):
        return self._add_obj(obj, ax_callback, plot_kwargs, as_image=True, single=True)
    
    def add_images(self, *objs, ax_callback=None, **plot_kwargs):
        for obj in objs:
            self.add_image(obj, ax_callback=ax_callback, **plot_kwargs)
        return self
    
    def add_plot(self, obj=None, ax_callback=None, **plot_kwargs):
        return self._add_obj(obj, ax_callback, plot_kwargs, as_image=False, single=True)
    
    def add_plots(self, *objs, ax_callback=None, **plot_kwargs):
        for obj in objs:
            self.add_plot(obj, ax_callback=ax_callback, **plot_kwargs)
        return self
    
    # endregion
    
    @staticmethod
    def get_grid_dims(obj_count, nrows=None, ncols=None):
        col_count = np.sqrt(obj_count * 16 / 9)
        if nrows is None:
            if ncols:
                nrows = max(1, np.ceil(obj_count / ncols))
            else:
                nrows = max(1, np.round(obj_count / col_count))
        if ncols is None:
            ncols = np.ceil(obj_count / nrows)
        
        pair = list(map(int, [nrows, ncols]))
        # logger.debug('nrows,ncols for %s images: %s',images_count,pair)
        return pair
    
    def plot(self, overlay=False, nrows_cols=None, horizontal=True, lbrtwh=(), **subplot_kwargs):
        if nrows_cols is None:
            if overlay:
                nrows, ncols = 1, 1
            else:
                nrows = subplot_kwargs.pop('nrows', None)
                ncols = subplot_kwargs.pop('ncols', None)
                nrows, ncols = self.get_grid_dims(len(self.data), nrows=nrows, ncols=ncols)
                if len(self.data) > nrows * ncols:
                    raise ValueError(
                        f'Insufficient number of axes ({nrows * ncols}) for {len(self.data)} objects.')
        else:
            nrows, ncols = nrows_cols
        fig, axs = plt.subplots(nrows, ncols, **subplot_kwargs)
        if nrows * ncols == 1:
            axs = [axs]
        else:
            if horizontal:
                axs = axs.flatten()
            else:
                axs = axs.T.flatten()
        
        for ax, datum in it.zip_longest(axs, self.data, fillvalue=axs[0] if overlay else None):
            if self.axis_off: ax.set_axis_off()
            if datum is None:
                fig.delaxes(ax)
                continue
            obj = datum.obj
            if obj is not None:
                if datum.as_image:
                    if datum.single:
                        assert isinstance(datum.plot_kwargs, dict)
                        ax.imshow(obj, **datum.plot_kwargs)
                    else:
                        objs = obj
                        if isinstance(datum.plot_kwargs, dict): datum.plot_kwargs = [datum.plot_kwargs]
                        for obj, plot_kwargs in it.zip_longest(objs, datum.plot_kwargs,
                                                               fillvalue=datum.plot_kwargs[-1]):
                            ax.imshow(obj, **plot_kwargs)
                else:
                    if datum.single:
                        if len(obj.shape) == 2:
                            assert isinstance(datum.plot_kwargs, dict)
                            ax.plot(*obj, **datum.plot_kwargs)
                        else:
                            ax.plot(obj, **datum.plot_kwargs)
                    else:
                        objs = obj
                        if isinstance(datum.plot_kwargs, dict): datum.plot_kwargs = [datum.plot_kwargs]
                        for obj, plot_kwargs in it.zip_longest(objs, datum.plot_kwargs,
                                                               fillvalue=datum.plot_kwargs[-1]):
                            if len(obj.shape) == 2:
                                ax.plot(*obj, **plot_kwargs)
                            else:
                                ax.plot(obj, **plot_kwargs)
            
            if datum.ax_callback is not None: datum.ax_callback(ax)
        
        fig.set_size_inches(18, 10)
        if lbrtwh:
            plt.subplots_adjust(*lbrtwh)
        return self
    
    @staticmethod
    def show(*args, **kwargs):
        return plt.show(*args, **kwargs)
    
    def save(self, savepath, *args, **kwargs):
        plt.savefig(savepath, *args, **kwargs)
        plt.close()
        return self
    
    def __len__(self):
        return len(self.data)


def main():
    pass


if __name__ == '__main__':
    main()



import wrapt

#
# basic_excludes=(int,float,str,bool)
#
# def _wrap_make_out():
# 	@wrapt.decorator
# 	def wrapper(wrapped,instance,args,kwargs):
# 		return instance.make_out(wrapped(*args,**kwargs))
#
# 	return wrapper
#
# # noinspection PyAbstractClass
# class RecursiveProxy(wrapt.ObjectProxy):
# 	exclude=()
# 	wrap_only=()
#
# 	def __init__(self,wrapped,*setup_args,**setup_kwargs):
# 		super().__init__(wrapped)
# 		self._self_exclude=tuple({*basic_excludes,*self.exclude})
# 		self._self_wrap_only=tuple(self.wrap_only)
# 		self._proxy_setup(*setup_args,**setup_kwargs)
#
# 	def _proxy_setup(self,*args,**kwargs):
# 		pass
#
# 	def __call__(self,*args,**kwargs):
# 		return self.__make_out(self.__wrapped__.__call__(*args,**kwargs))
#
# 	def make_out(self,out):
# 		return out
#
# 	def inherit_out(self,out):
# 		return type(self)(out)
#
# 	@_wrap_make_out()
# 	def __make_out(self,out):
# 		if self._self_wrap_only:
# 			if isinstance(out,self._self_wrap_only):
# 				return self.inherit_out(out)
# 			else:
# 				return out
# 		elif isinstance(out,self._self_exclude):
# 			return out
# 		else:
# 			return self.inherit_out(out)
#
# 	def __getattr__(self,item):
# 		if item.startswith('_self_'):
# 			return super().__getattr__(item)
# 		else:
# 			return self.__make_out(getattr(self.__wrapped__,item))
#
# 	@property
# 	def the_obj(self):
# 		return self.__wrapped__
#
# 	def __reduce_ex__(self,protocol):
# 		args=(self.__wrapped__,)
# 		return type(self),args

basic_excludes = (int, float, str, bool)


def _wrap_make_out():
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        return instance.make_out(wrapped(*args, **kwargs))
    
    return wrapper


def _new_obj_proxy(cls, args, kwargs):
    # logger.warning('_new_obj_proxy')
    # logger.debug('cls: %s',cls)
    # args,kwargs=args_kwargs
    # logger.debug('args: %s',args)
    # logger.debug('kwargs: %s',kwargs)
    "A function to map kwargs into cls.__new__"
    # return cls.__new__(cls,*args,**kwargs)
    return cls(*args, **kwargs)


# noinspection PyAbstractClass
class RecursiveProxy(wrapt.ObjectProxy):
    _exclude = ()
    _wrap_only = ()
    
    def __init__(self, wrapped, *setup_args, **setup_kwargs):
        super().__init__(wrapped)
        self._self_setup_args = setup_args
        self._self_setup_kwargs = setup_kwargs
        self._self_exclude = tuple({*basic_excludes, *self._exclude})
        self._self_wrap_only = tuple(self._wrap_only)
        self._proxy_setup(*setup_args, **setup_kwargs)
    
    def _proxy_setup(self, *args, **kwargs):
        pass
    
    def __call__(self, *args, **kwargs):
        return self._make_out(self.__wrapped__.__call__(*args, **kwargs))
    
    # def make_out(self,out):
    # 	return out
    
    def _make_out(self, out):
        if self._self_wrap_only:
            if isinstance(out, self._self_wrap_only):
                return type(self)(out, *self._self_setup_args, **self._self_setup_kwargs)
            else:
                return out
        elif isinstance(out, self._self_exclude):
            return out
        else:
            return type(self)(out, *self._self_setup_args, **self._self_setup_kwargs)
    
    def __getattr__(self, item):
        if item.startswith('_self_'):
            return super().__getattr__(item)
        else:
            return self._make_out(getattr(self.__wrapped__, item))
    
    @property
    def _class_(self):
        return type(self)
    
    @property
    def the_obj(self):
        return self.__wrapped__
    
    def __reduce_ex__(self, protocol):
        args = (self.__wrapped__, *self._self_setup_args)
        return _new_obj_proxy, (type(self), args, self._self_setup_kwargs)
    
    def __getitem__(self, item):
        return self._make_out(super().__getitem__(item))



import geojson
import numpy as np
from box import Box
from lgblkb_tools import logger
from matplotlib import pyplot as plt
from matplotlib.path import Path
from numpy import ones, concatenate, asarray
from shapely import geometry as shg


def ring_coding(ob):
    # The codes will be all "LINETO" commands, except for "MOVETO"s at the
    # beginning of each subpath
    n = len(ob.coords)
    codes = ones(n, dtype=Path.code_type) * Path.LINETO
    codes[0] = Path.MOVETO
    return codes


def pathify(polygon):
    # Convert coordinates to path vertices. Objects produced by Shapely's
    # analytic methods have the proper coordinate order, no need to sort.
    vertices = concatenate(
        [asarray(polygon.exterior)]
        + [asarray(r) for r in polygon.interiors])
    codes = concatenate(
        [ring_coding(polygon.exterior)]
        + [ring_coding(r) for r in polygon.interiors])
    return Path(vertices, codes)


def perpendicular(a):
    b = np.empty_like(a)
    b[0] = -a[1]
    b[1] = a[0]
    return b


def normalize(a):
    a = np.array(a)
    return a / np.linalg.norm(a)


def plot_line(line):
    return plt.plot(*line.xy, lw=5)


round_decimals = 6
min_dist = 1e-6


class GenericGeometry(object):

    def __init__(self, **data):
        self.data = Box(data, ordered_box=True)

    # self.id=uuid.uuid4()

    # def _generate_id(self,id_obj=None):
    # 	self.id=self.data.pop('id',gsup.get_md5(str(id_obj or self.geometry)))

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.geometry}"

    def __getitem__(self, item):
        # logger.debug('item: %s',item)
        return self.data[item]
        # raise KeyError('Invalid key provided.',dict(key=item,key_type=type(item)))
        pass

    def __hash__(self):
        return hash(self.geometry.wkt)

    @property
    def geometry(self):
        raise NotImplementedError

    def almost_touches(self, other):
        return self.geometry.distance(other.geometry) < min_dist


def line_xy(linestring):
    return np.array(linestring.xy).T


def get_rounded(linestring, decimals):
    new_linestring = shg.LineString(np.round(line_xy(linestring), decimals=decimals))
    return new_linestring


def get_lines(*polygons):
    lines = list()
    for pol in [polygons]:
        boundary = pol.boundary
        if boundary.type == 'MultiLineString':
            for line in boundary:
                lines.append(line)
        else:
            lines.append(boundary)
    return lines


def coors_to_img(coors):
    # get ranges
    xmin = coors[:, 0].min()
    xmax = coors[:, 0].max()
    ymin = coors[:, 1].min()
    ymax = coors[:, 1].max()
    zmin = coors[:, 2].min()
    zmax = coors[:, 2].max()

    # create array for image : zmax+1 is the default value
    shape = (xmax - xmin + 1, ymax - ymin + 1)
    img = np.ma.array(np.ones(shape) * (zmax + 1))

    for inp in coors:
        img[inp[0] - xmin, inp[1] - ymin] = inp[2]

    # set mask on default value
    img.mask = (img == zmax + 1)

    # set a gray background for test
    img_bg_test = np.zeros(shape)
    cmap_bg_test = plt.get_cmap('gray')
    plt.imshow(img_bg_test, cmap=cmap_bg_test, interpolation='none')

    # plot
    cmap = plt.get_cmap('jet')
    plt.imshow(img, cmap=cmap, interpolation='none', vmin=zmin, vmax=zmax)
    plt.colorbar()

    plt.imsave("test.png", img)
    plt.show()
    plt.close()
    pass


def is_clockwise(polygon: shg.Polygon):
    xy = line_xy(polygon.boundary)
    x = xy[:, 0]
    y = xy[:, 1]
    diff_x = x[1:] - x[:-1]
    sum_y = y[:-1] + y[1:]
    area = np.sum(diff_x * sum_y)
    # logger.info('x:\n%s',x)
    # logger.info('y:\n%s',y)
    # logger.info('diff_x:\n%s',diff_x)
    # logger.info('sum_y:\n%s',sum_y)
    # logger.info('area: %s',area)
    if area > 0:
        __isclockwise = True
    elif area < 0:
        __isclockwise = False
    else:
        logger.info('self.polygon:\n%s', polygon)
        raise NotImplementedError('Area is zero.')
    return __isclockwise


def cut_polygon(polygon, cut_line):
    mpoly = polygon.difference(cut_line.buffer(1e-9))
    # logger.info('mpoly:\n%s',mpoly)
    return mpoly


def generate_cells(polygon, cutlines):
    if not cutlines: return [polygon]
    cells = list()
    # for i,cutline in enumerate(cutlines):
    if not polygon.intersects(cutlines[0]): return [polygon]
    mpoly = cut_polygon(polygon, cutlines[0])
    assert len(mpoly) > 1
    for p in mpoly:
        cells.extend(generate_cells(p, cutlines[1:]))
    return cells


def get_from_geojson(filepath):
    geojson_data = geojson.load(open(filepath), )
    geoms = [shg.shape(x['geometry']).buffer(0) for x in geojson_data['features']]
    return geoms


def main():
    pass


if __name__ == '__main__':
    main()



import pandas as pd
import itertools as it
import collections
import functools
import hashlib
import multiprocessing as mp
import os
import signal
import subprocess
import sys
import time
from datetime import datetime, timedelta, date
from typing import Sized

import dateutil
import dateutil.parser

from .. import logger


def reprer(obj):
    if isinstance(obj, dict):
        d = obj
    else:
        d = obj.__dict__
    body = [f"{k:<20}: {('None' if v is None else v)}" for k, v in d.items() if k[0] != '_']
    if body:
        delim = "\n"
    # else: body,delim=["Expired!"]," "
    else:
        body, delim = ['Empty'], " "
    return f'"{obj.__class__.__name__}" item:' + delim + "\n".join(body)


def func_has_classarg(func, args):
    has_classarg = False
    if args:
        funcattr = getattr(args[0], func.__name__, None)
        if funcattr is not None and hasattr(funcattr, '__self__'):
            has_classarg = True
    return has_classarg


def infiterate(iter_obj, max_iter_count=None, next_getter=None, inform_count=8, on_inform=None):
    if on_inform is None: on_inform = lambda *args, **kwargs: True
    if isinstance(iter_obj, int):
        an_iterable, iter_count = range(iter_obj), iter_obj
    else:
        an_iterable = iter_obj
        if isinstance(iter_obj, Sized):
            iter_count = len(iter_obj)
        else:
            assert max_iter_count
            iter_count = max_iter_count
    max_iter_count = max_iter_count or iter_count
    if next_getter is None: next_getter = lambda _obj: _obj
    for i, obj in enumerate(an_iterable):
        if i % (max_iter_count // inform_count) == 0:
            on_inform()
            logger.info('%d%%, i=%d', i / (max_iter_count - 1) * 100, i)

        obj = next_getter(obj)
        yield obj
        if i == max_iter_count - 1:
            on_inform()
            logger.info('%d%%, i=%d', i / (max_iter_count - 1) * 100, i)
            return


class ParallelTasker:

    def __init__(self, func, *args, **kwargs):
        self.partial_func = functools.partial(func, *args, **kwargs)
        self.queue = mp.Queue()
        self._total_proc_count = 0
        pass

    def set_run_params(self, **kwargs):
        val_lengths = {len(v) for v in kwargs.values()}
        assert len(val_lengths) == 1
        val_length = val_lengths.pop()

        for i in range(val_length):
            self.queue.put((i, {k: v[i] for k, v in kwargs.items()}))
            self._total_proc_count += 1
        # simple_logger.info('self._total_proc_count: %s',self._total_proc_count)

        return self

    def __process_func(self, queue, common_list):
        i, kwargs = queue.get()
        result = self.partial_func(**kwargs)
        common_list.append([i, result])

    @staticmethod
    def __join_finished_processes(active_procs, sleep_time):
        while True:
            # simple_logger.info('Process queue is full. Searching for finished processes.')
            for p in active_procs:
                if not p.is_alive():
                    # simple_logger.info('Finished process found. Joining it.')
                    active_procs.remove(p)
                    p.join()
                    # p.terminate()
                    # simple_logger.info('Process successfully removed.')
                    return
            time.sleep(sleep_time)

    def run(self, workers_count=None, sleep_time=1.0):
        workers_count = min(mp.cpu_count() - 1, workers_count or self._total_proc_count)
        manager = mp.Manager()
        common_list = manager.list()
        processes = [mp.Process(target=self.__process_func, args=(self.queue, common_list)) for _ in
                     range(self.queue.qsize())]
        active_procs = list()
        while True:
            # simple_logger.info('loop begins')
            if len(processes) == 0:
                # simple_logger.info('Waiting for the last jobs to finish.')
                for active_p in active_procs:
                    active_p.join()
                break
            if len(active_procs) < workers_count:
                # simple_logger.info('Adding a process')
                proc = processes.pop()
                proc.start()
                active_procs.append(proc)
            else:
                self.__join_finished_processes(active_procs, sleep_time)
            time.sleep(sleep_time)
        return [item[1] for item in sorted(common_list, key=lambda x: x[0])]


def run_shell(*args, non_blocking=False, chdir=None, with_popen=False, ret_triggers=None, **kwargs):
    chdir = chdir or os.getcwd()
    normal_dir = os.getcwd()
    os.chdir(chdir)

    if non_blocking:
        subprocess.Popen(args, stdout=subprocess.PIPE, **kwargs)
    else:
        # output=subprocess.run(args,stdout=subprocess.PIPE,**kwargs).stdout.decode()
        # output=subprocess.run(args,stdout=subprocess.PIPE,**kwargs).stdout.decode()
        if with_popen:
            process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, preexec_fn=os.setsid)
            regular_termination = True
            nextline = ''
            while regular_termination:
                nextline = process.stdout.readline().decode()
                # print('nextline:',nextline)
                if nextline == '' and process.poll() is not None:
                    break
                if ret_triggers:
                    for ret_trigger in ret_triggers:
                        if ret_trigger in nextline:
                            regular_termination = False
                            break
                logger.debug(nextline)
                # sys.stdout.write(nextline)
                sys.stdout.flush()
            if regular_termination:
                output = process.communicate()[0]
                exitCode = process.returncode
                if exitCode != 0:
                    raise Exception(args, exitCode, output)
            else:
                logger.info(r'Manual process termination triggered with line:\n%s', nextline)
                logger.debug('Killing process: %s', process.pid)
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                raise Exception('Manual process termination.')
        else:
            subprocess.run(args, stdout=subprocess.PIPE, **kwargs).stdout.decode()
    os.chdir(normal_dir)


def period_within(days=None, start_date=None, end_date=None):
    if isinstance(start_date, str): start_date = dateutil.parser.parse(start_date)
    if isinstance(end_date, str): end_date = dateutil.parser.parse(end_date)
    if days is not None:
        if start_date is not None:
            end_date = start_date + timedelta(days=days)
        elif end_date is not None:
            start_date = end_date - timedelta(days=days)
        else:
            end_date = date.today()
            start_date = end_date - timedelta(days=days)
    else:
        end_date = end_date or date.today()
    return start_date, end_date


def datetime_within(start=None, end=None, **timedelta_opts):
    if timedelta_opts:
        if start is not None:
            end = start + timedelta(**timedelta_opts)
        elif end is not None:
            start = end - timedelta(**timedelta_opts)
        else:
            end = datetime.now()
            start = end - timedelta(**timedelta_opts)
    else:
        end = end or datetime.now()
    return start, end


def md5(filepath):
    hash_md5 = hashlib.md5()
    with open(filepath, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_md5(text):
    return hashlib.md5(str(text).encode('utf-8')).hexdigest()


def run_command(cmd):
    logger.debug('cmd: %s', cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    for c in iter(lambda: process.stdout.read(1), b''):  # replace '' with b'' for Python 3
        sys.stdout.write(c.decode(sys.stdout.encoding, errors='ignore'))


def run_cmd(cmd, debug=False, **kwargs):
    if isinstance(cmd, str):
        steps = [cmd]
    else:
        steps = cmd
    for step in steps:
        if debug: logger.debug('step: %s', step)
        subprocess.run(step, **dict(dict(check=True, shell=True), **kwargs))


def dict_merge(dct, merge_dct, add_keys=True):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param add_keys: add missing keys from merge_dct to dct.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    # for k,v in merge_dct.iteritems():
    # 	if (k in dct and isinstance(dct[k],dict)
    # 			and isinstance(merge_dct[k],collections.Mapping)):
    # 		dict_merge(dct[k],merge_dct[k])
    # 	else:
    # 		dct[k]=merge_dct[k]

    for k, v in merge_dct.items():
        if isinstance(dct.get(k), dict) and isinstance(v, collections.Mapping):
            dct[k] = dict_merge(dct[k], v, add_keys=add_keys)
        else:
            dct[k] = v
    return dct


def get_dist_info(df, dists):
    """
    
    :param df: any object having shape attribute (e.g. pandas Dataframe, numpy array).
    :param dists: the output of from "scipy.spatial.distance.pdist" function.
    :return: pandas Dataframe with ['pair','dist'] columns representing the row indices and distances between them.
    """
    dists_info = pd.DataFrame(zip(it.combinations(range(df.shape[0]), 2), dists), columns=['pair', 'dist'])
    return dists_info



from pathlib import Path
import logging
import os
import string
import uuid
from typing import Iterable, Union

import geojson
import matplotlib.pyplot as plt
import numpy as np
import shapely.geometry as shg
import shapely.wkt as shwkt
from box import Box
from osgeo import gdal, gdal_array, ogr, osr, gdalconst

from . import logger
from . import pathify
from .common.utils import run_cmd
from .pathify import get_name

gdal.UseExceptions()


def ds_to_array(ds, band_index=1):
    if isinstance(ds, str): ds = gdal.Open(ds)
    return ds.GetRasterBand(band_index).ReadAsArray(0, 0, ds.RasterXSize, ds.RasterYSize)


geom_from_ds = lambda ds: np.abs(np.array(ds.GetGeoTransform()).reshape(2, -1))
resol_from_ds = lambda ds: geom_from_ds(ds)[:, 1:].sum(axis=0)
ul_from_ds = lambda ds: geom_from_ds(ds)[:, 0]


# get_epsg_from = lambda ds: int(
#     gdal.Info(ds, format='json')['coordinateSystem']['wkt'].rsplit('"EPSG","', 1)[-1].split('"')[0])


def plot_ds(ds, show=False):
    array = ds_to_array(ds)
    plt.imshow(array)
    if show: plt.show()


code_lut = {v: k for k, v in gdal_array.codes.items()}


def array_to_ds(array, parent_info):
    try:
        arr_ds = gdal_array.OpenNumPyArray(array, True)
    except:
        arr_ds = gdal_array.OpenNumPyArray(array)
    if isinstance(parent_info, DataSet): parent_info = parent_info.ds
    if isinstance(parent_info, gdal.Dataset):
        transform = parent_info.GetGeoTransform()
        projection = parent_info.GetProjection()
    else:
        transform, projection = parent_info
    arr_ds.SetGeoTransform(transform)
    arr_ds.SetProjection(projection)
    return arr_ds


def get_datasource(driver_name, fpath, field_name, field_type):
    dest_datasource = ogr.GetDriverByName(driver_name).CreateDataSource(fpath)
    dest_layer = dest_datasource.CreateLayer('', srs=None)
    dest_layer.CreateField(ogr.FieldDefn(field_name, field_type))
    return dest_datasource, dest_layer


@logger.trace()
def polygonize(ds, driver_name, fpath, field_name='DN', field_type=ogr.OFTInteger):
    dest_datasource, dest_layer = get_datasource(driver_name, fpath, field_name, field_type)
    if field_type is ogr.OFTReal:
        polygonizer = gdal.FPolygonize
    else:
        polygonizer = gdal.Polygonize
    polygonizer(ds.GetRasterBand(1), None, dest_layer, 0, ['8connected=8'], callback=None)
    return dest_datasource, dest_layer


def contour_generate(ds, val_ranges, driver_name, fpath, field_name, field_type):
    contour_dsrc, contour_layer = get_datasource(driver_name, fpath, field_name, field_type)
    gdal.ContourGenerate(ds.GetRasterBand(1), 0, 0, val_ranges, 0, 0, contour_layer, 0, 0)
    return contour_dsrc, contour_layer


class DataSet:

    def __init__(self, ds, array=None, name=''):
        self.path = None
        if isinstance(ds, str):
            name = name or '_'.join(pathify.get_splitted(os.path.splitext(ds)[0])[-2:])
            self.path = ds
            ds = gdal.Open(ds)
        self.ds = ds
        geom = geom_from_ds(self.ds)
        self.resol = geom[:, 1:].sum(axis=0)
        self.ul = geom[:, 0]
        self.__array = array
        self.projection = self.ds.GetProjection()
        self.transform = self.ds.GetGeoTransform()
        self.name = name

    # @logger.trace(skimpy=True)
    def get_array(self, band_index=1):
        return ds_to_array(self.ds, band_index=band_index)

    @property
    def array(self):
        if self.__array is None:
            self.__array = self.get_array()
        return self.__array

    @property
    def epsg(self):
        return osr.SpatialReference(wkt=self.projection).GetAttrValue('AUTHORITY', 1)

    def plot(self, show=False):
        plot_ds(self.ds, show=show)
        return self

    @logger.trace()
    def reproject(self, to_epsg):
        """
        A sample function to reproject and resample a GDAL dataset from within
        Python. The idea here is to reproject from one system to another, as well
        as to change the pixel size. The procedure is slightly long-winded, but
        goes like this:

        1. Set up the two Spatial Reference systems.
        2. Open the original dataset, and get the geotransform
        3. Calculate bounds of new geotransform by projecting the UL corners
        4. Calculate the number of pixels with the new projection & spacing
        5. Create an in-memory raster dataset
        6. Perform the projection
        """
        # Define the UK OSNG, see <http://spatialreference.org/ref/epsg/27700/>
        osng = osr.SpatialReference()
        osng.ImportFromEPSG(to_epsg)
        wgs84 = osr.SpatialReference()
        wgs84.ImportFromEPSG(self.epsg)
        tx = osr.CoordinateTransformation(wgs84, osng)
        # Up to here, all  the projection have been defined, as well as a
        # transformation from the from to the  to :)
        # We now open the dataset
        # Get the Geotransform vector
        geo_t = self.ds.GetGeoTransform()
        x_size, y_size = self.raster_sizes
        # Work out the boundaries of the new dataset in the target projection
        (ulx, uly, ulz) = tx.TransformPoint(geo_t[0], geo_t[3])
        (lrx, lry, lrz) = tx.TransformPoint(geo_t[0] + geo_t[1] * x_size, geo_t[3] + geo_t[5] * y_size)
        # See how using 27700 and WGS84 introduces a z-value!
        # Now, we create an in-memory raster
        mem_drv = gdal.GetDriverByName('MEM')
        # The size of the raster is given the new projection and pixel spacing
        # Using the values we calculated above. Also, setting it to store one band
        # and to use Float32 data type.
        # dest=mem_drv.Create('',int((lrx-ulx)/x_size),int((uly-lry)/y_size),1,gdal.GDT_Float32)
        dtype = np.typeDict[str(self.array.dtype)]
        dest = mem_drv.Create('', x_size, y_size, 1, code_lut[dtype])
        # dest=mem_drv.Create('',pixel_x_size,pixel_y_size,1,gdal.GDT_Float32)
        # Calculate the new geotransform
        resol = self.resol
        new_geo = (ulx, resol[0], geo_t[2], uly, geo_t[4], -resol[1])
        # Perform the projection/resampling
        # Set the geotransform
        dest.SetGeoTransform(new_geo)
        dest.SetProjection(osng.ExportToWkt())
        gdal.ReprojectImage(self.ds, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Bilinear)
        # gdal.ReprojectImage(self.ds,dest,wgs84.ExportToWkt(),osng.ExportToWkt())
        return DataSet(dest)

    # @logger.trace(skimpy=True)
    def do_warp(self, cutline_feature=None, destroy_after=True, out_epsg=None, out_path='', return_as_ds=True,
                lazy=False, ignore_error=False, **kwargs):
        """options=None, format=None,
         outputBounds=None,
         outputBoundsSRS=None,
         xRes=None, yRes=None, targetAlignedPixels = False,
         width = 0, height = 0,
         srcSRS=None, dstSRS=None,
         srcAlpha = False, dstAlpha = False,
         warpOptions=None, errorThreshold=None,
         warpMemoryLimit=None, creationOptions=None, outputType = gdalconst.GDT_Unknown,
         workingType = gdalconst.GDT_Unknown, resampleAlg=None,
         srcNodata=None, dstNodata=None, multithread = False,
         tps = False, rpc = False, geoloc = False, polynomialOrder=None,
         transformerOptions=None, cutlineDSName=None,
         cutlineLayer=None, cutlineWhere=None, cutlineSQL=None, cutlineBlend=None, cropToCutline = False,
         copyMetadata = True, metadataConflictValue=None,
         setColorInterpretation = False,
         callback=None, callback_data=None"""
        if lazy and out_path and (not return_as_ds) and os.path.exists(out_path): return out_path
        out_path = out_path or f"/vsimem/{uuid.uuid4()}"
        if not cutline_feature is None:
            if isinstance(cutline_feature, str):
                cutline_feature = geojson.Feature(geometry=shwkt.loads(cutline_feature))
            elif isinstance(cutline_feature, (shg.MultiPolygon, shg.Polygon)):
                cutline_feature = geojson.Feature(geometry=cutline_feature)
            elif isinstance(cutline_feature, DataSet):
                cutline_feature = geojson.Feature(geometry=cutline_feature.geom)
        try:
            gdal.Warp(out_path, self.ds, srcSRS=f'EPSG:{self.epsg}', dstSRS=f'EPSG:{out_epsg or self.epsg}',
                      cutlineDSName=cutline_feature, **dict(dict(cropToCutline=not cutline_feature is None), **kwargs))
        except RuntimeError as exc:
            if return_as_ds or not ignore_error: raise exc
            logger.warning(str(exc))
            return
        if return_as_ds:
            out = DataSet(out_path)
        else:
            out = out_path
        if destroy_after: self.ds = None
        return out

    @property
    def geo_info(self):
        return self.transform, self.projection

    @property
    def raster_count(self):
        return self.ds.RasterCount

    @property
    def raster_sizes(self):
        return self.ds.RasterXSize, self.ds.RasterYSize

    @property
    def file_meta(self):
        return Box(self.ds.GetMetadata())

    def get_band_metas(self):
        band_metas = list()
        for i in self.raster_count:
            band_metas.append(Box(self.ds.GetRasterBand(i + 1).GetMetadata()))
        return band_metas

    @classmethod
    def from_array(cls, array, geo_info):
        if isinstance(geo_info, str): geo_info = DataSet(geo_info)
        return DataSet(array_to_ds(array, geo_info), array)

    @logger.trace()
    def polygonize(self, driver_name, fpath, destroy_after=True, **kwargs):
        datasource, ds_layer = polygonize(self.ds, driver_name, fpath, **kwargs)
        if not 'mem' in driver_name.lower():
            ds_layer = None
            datasource = None
        if destroy_after: self.ds = None
        return datasource, ds_layer

    @logger.trace()
    def generate_contours(self, val_ranges, driver_name, fpath, field_name='DN',
                          field_type=ogr.OFTReal, destroy_after=True):
        datasource, ds_layer = contour_generate(ds=self.ds, val_ranges=val_ranges,
                                                driver_name=driver_name, fpath=fpath,
                                                field_name=field_name, field_type=field_type)
        if not 'mem' in driver_name.lower():
            ds_layer = None
            datasource = None
        if destroy_after: self.ds = None
        return datasource, ds_layer

    @logger.trace(skimpy=True)
    def to_geotiff(self, filepath, no_data_value=-9999, dtype=gdal.GDT_Float32, lazy=False):
        if lazy and os.path.exists(filepath): return filepath
        band = self.ds.GetRasterBand(1)
        arr = band.ReadAsArray()
        [cols, rows] = arr.shape
        driver = gdal.GetDriverByName("GTiff")
        outdata = driver.Create(filepath, rows, cols, 1, dtype)
        outdata.SetGeoTransform(self.transform)  ##sets same geotransform as input
        outdata.SetProjection(self.projection)  ##sets same projection as input
        outdata.GetRasterBand(1).WriteArray(np.where(np.isnan(arr), no_data_value, arr))
        if no_data_value is not False:
            outdata.GetRasterBand(1).SetNoDataValue(no_data_value)  ##if you want these values transparent
        outdata.FlushCache()  ##saves to disk!!
        outdata = None
        band = None
        ds = None
        return filepath

    @logger.trace(skimpy=True)
    def to_file(self, filepath, driver_name, no_data_value=-9999, dtype=gdal.GDT_Float32,
                file_meta: dict = None, band_metas: list = None):
        driver = gdal.GetDriverByName(driver_name)
        cols, rows = self.ds.GetRasterBand(1).ReadAsArray().shape
        outdata = driver.Create(filepath, rows, cols, self.raster_count, dtype)
        outdata.SetGeoTransform(self.transform)  ##sets same geotransform as input
        outdata.SetProjection(self.projection)  ##sets same projection as input
        if file_meta:
            outdata.SetMetadata({str(k): str(v) for k, v in file_meta.items()})

        for band_i in range(self.ds.RasterCount):
            band_num = band_i + 1
            arr = self.ds.GetRasterBand(band_num).ReadAsArray()
            if no_data_value is False:
                the_array = arr
            else:
                the_array = np.where(np.isnan(arr), no_data_value, arr)
            band = outdata.GetRasterBand(band_num)
            band.WriteArray(the_array)
            if band_metas:
                band.SetMetadata({str(k): str(v) for k, v in band_metas[band_i].items()})
            if no_data_value is not False:
                band.SetNoDataValue(no_data_value)  ##if you want these values transparent
        outdata.FlushCache()  ##saves to disk!!
        outdata = None
        band = None
        return filepath

    @logger.trace(skimpy=True)
    def to_file_old(self, filepath, driver_name, no_data_value=-9999, dtype=gdal.GDT_Float32):
        band = self.ds.GetRasterBand(1)
        arr = band.ReadAsArray()
        [cols, rows] = arr.shape
        driver = gdal.GetDriverByName(driver_name)
        outdata = driver.Create(filepath, rows, cols, 1, dtype)
        outdata.SetGeoTransform(self.transform)  ##sets same geotransform as input
        outdata.SetProjection(self.projection)  ##sets same projection as input
        outdata.GetRasterBand(1).WriteArray(np.where(np.isnan(arr), no_data_value, arr))
        if no_data_value is not False:
            outdata.GetRasterBand(1).SetNoDataValue(no_data_value)  ##if you want these values transparent
        outdata.FlushCache()  ##saves to disk!!
        outdata = None
        band = None
        ds = None
        return filepath

    def scale_array(self, scaler):
        scaled = scaler.fit_transform(self.array.reshape(-1, 1))
        scaled_band = scaled.reshape(self.array.shape)
        out = self.from_array(scaled_band, self.geo_info)
        self.ds = None
        return out

    def get_info(self, **kwargs):
        kwargs = dict(dict(format='json'), **kwargs)
        return gdal.Info(self.ds, options=gdal.InfoOptions(**kwargs))

    @logger.trace(skimpy=True)
    def translate(self, options=None, format=None,
                  outputType=gdalconst.GDT_Unknown, bandList=None, maskBand=None,
                  width=0, height=0, widthPct=0.0, heightPct=0.0,
                  xRes=0.0, yRes=0.0,
                  creationOptions=None, srcWin=None, projWin=None, projWinSRS=None, strict=False,
                  unscale=False, scaleParams=None, exponents=None,
                  outputBounds=None, metadataOptions=None,
                  outputSRS=None, GCPs=None,
                  noData=None, rgbExpand=None,
                  stats=False, rat=True, resampleAlg=None,
                  callback=None, callback_data=None, destroy_after=True, out_path='', return_as_ds=True):
        out_path = out_path or f"/vsimem/{uuid.uuid4()}"
        gdal.Translate(out_path, self.ds, options=gdal.TranslateOptions(
            options=options, format=format, outputType=outputType, bandList=bandList, maskBand=maskBand, width=width,
            height=height,
            widthPct=widthPct, heightPct=heightPct, xRes=xRes, yRes=yRes, creationOptions=creationOptions,
            srcWin=srcWin, projWin=projWin,
            projWinSRS=projWinSRS, strict=strict, unscale=unscale, scaleParams=scaleParams, exponents=exponents,
            outputBounds=outputBounds,
            metadataOptions=metadataOptions, outputSRS=outputSRS, GCPs=GCPs, noData=noData, rgbExpand=rgbExpand,
            stats=stats, rat=rat,
            resampleAlg=resampleAlg, callback=callback, callback_data=callback_data,
        ))
        if return_as_ds:
            out = DataSet(out_path, name='translated')
        else:
            out = out_path
        if destroy_after: self.ds = None
        return out

    @property
    def wkt(self):
        return self.geom.wkt

    @property
    def geom(self):
        return shg.shape(self.get_info()['wgs84Extent'])


@logger.trace(level=logging.DEBUG)
def translate_tiff(input_tiff, output_tiff, translate_opts: gdal.TranslateOptions) -> str:
    ds = gdal.Open(input_tiff)
    ds = gdal.Translate(output_tiff, ds, options=translate_opts)
    ds = None
    return output_tiff


@logger.trace(level=logging.DEBUG)
def rgb_to_geotiff(tiff_path, *band_paths, no_data_value=-9999, dtype=gdal.GDT_Float32, **warp_opts):
    # logsup.logger.info('no_data_value: %s',no_data_value)
    band_dss = [DataSet(path).do_warp(**warp_opts) for path in band_paths]
    band_arrays = [band_ds.array for band_ds in band_dss]
    # for path in band_paths:
    # 	step1=DataSet(path).do_warp(out_epsg=3857,srcNodata=-2000)  #
    # 	step3=step1.scale_array(RobustScaler()).scale_array(MinMaxScaler((0,255))).array
    # 	band_arrays.append(step3)
    # band_arrays=[_get_image_prepared_array(path,warp_opts=dict(out_epsg=3857),scaler=MinMaxScaler((0,255))) for path in band_paths]
    # band_arrays=gsup.ParallelTasker(_get_image_prepared_array,scaler=MinMaxScaler((0,255))).set_run_params(ds=band_paths).run(sleep_time=1e-1)
    # band_arrays=gsup.ParallelTasker(ds_to_array).set_run_params(ds=band_paths).run(sleep_time=1e-1)
    [cols, rows] = band_arrays[0].shape
    driver = gdal.GetDriverByName("GTiff")
    if len(band_paths) == 3:
        options = ['PHOTOMETRIC=RGB', 'PROFILE=GeoTIFF', ]
    else:
        options = None
    # elif len(band_paths)==4:
    # 	options=['PHOTOMETRIC=RGBA','PROFILE=GeoTIFF',]
    # else:
    # 	raise NotImplementedError(f'Invalid number of input files - {len(band_paths)}.',dict(count=len(band_paths)))

    outdata = driver.Create(tiff_path, rows, cols, len(band_paths), dtype, options=options)
    outdata.SetGeoTransform(band_dss[0].transform)  ##sets same geotransform as input
    outdata.SetProjection(band_dss[0].projection)  ##sets same projection as input
    for i, (array, band_interpretation) in \
            enumerate(zip(band_arrays, [gdal.GCI_RedBand, gdal.GCI_GreenBand, gdal.GCI_BlueBand])):
        raster_band = outdata.GetRasterBand(i + 1)
        raster_band.SetColorInterpretation(band_interpretation)
        raster_band.WriteArray(np.where(array == 0, no_data_value, array))
        if no_data_value is not False:
            raster_band.SetNoDataValue(no_data_value)  ##if you want these values transparent

    outdata.FlushCache()  ##saves to disk!!
    outdata = None
    band = None
    ds = None
    return tiff_path


def get_mean_std_scale_params(stack: np.array, std_num=2):
    mean = np.nanmean(stack)
    std = np.nanstd(stack)
    return [max(0, mean - std_num * std), mean + std_num * std]


def GetExtent(gt, cols, rows):
    ''' Return list of corner coordinates from a geotransform

        @type gt:   C{tuple/list}
        @param gt: geotransform
        @type cols:   C{int}
        @param cols: number of columns in the dataset
        @type rows:   C{int}
        @param rows: number of rows in the dataset
        @rtype:    C{[float,...,float]}
        @return:   coordinates of each corner
    '''
    ext = []
    xarr = [0, cols]
    yarr = [0, rows]

    for px in xarr:
        for py in yarr:
            x = gt[0] + (px * gt[1]) + (py * gt[2])
            y = gt[3] + (px * gt[4]) + (py * gt[5])
            ext.append([x, y])
        # print(x,y)

        yarr.reverse()
    return ext


def ReprojectCoords(coords, src_srs, tgt_srs):
    ''' Reproject a list of x,y coordinates.

        @type geom:     C{tuple/list}
        @param geom:    List of [[x,y],...[x,y]] coordinates
        @type src_srs:  C{osr.SpatialReference}
        @param src_srs: OSR SpatialReference object
        @type tgt_srs:  C{osr.SpatialReference}
        @param tgt_srs: OSR SpatialReference object
        @rtype:         C{tuple/list}
        @return:        List of transformed [[x,y],...[x,y]] coordinates
    '''
    trans_coords = []
    transform = osr.CoordinateTransformation(src_srs, tgt_srs)
    for x, y in coords:
        x, y, z = transform.TransformPoint(x, y)
        trans_coords.append([x, y])
    return trans_coords


def get_geo_extent(ds):
    gt = ds.GetGeoTransform()
    cols = ds.RasterXSize
    rows = ds.RasterYSize
    ext = GetExtent(gt, cols, rows)

    src_srs = osr.SpatialReference()
    src_srs.ImportFromWkt(ds.GetProjection())
    # tgt_srs=osr.SpatialReference()
    # tgt_srs.ImportFromEPSG(4326)
    tgt_srs = src_srs.CloneGeogCS()

    geo_ext = ReprojectCoords(ext, src_srs, tgt_srs)
    return geo_ext


class GdalMan(object):
    def __init__(self, debug=False, lazy=False, **kwargs):
        self.out_filepath = None
        self.kwargs = kwargs
        self.debug = debug
        self.lazy = lazy
        self.data = dict()

    def _get_laziness(self, lazy, out_filepath):
        # if Path(out_filepath).suffix.lower() == '.vrt' and lazy:
        #     logger.warning('Lazy is set to True, but provided out_filepath is virtual.',
        #                    dict(provided_vrt_path=out_filepath))
        #     lazy = False
        return self.lazy if lazy is None else lazy

    @property
    def path(self):
        return self.out_filepath

    @property
    def ds(self):
        return DataSet(self.out_filepath)

    def _run_gdal_cmd(self, gdal_func, *args, is_gdal_calc=False, **input_params):
        params = list()
        if is_gdal_calc:
            default_kwargs = self.kwargs.copy()
            if 'q' in self.kwargs:
                q = default_kwargs.pop('q')
                default_kwargs['--quiet'] = q
            options = dict(default_kwargs, **input_params)
        else:
            options = dict(self.kwargs, **input_params)
        for k, v in options.items():
            if v is True:
                if k.startswith('-'):
                    params.append(k)
                else:
                    params.append(f'-{k}')
            elif not v:
                continue
            else:
                if k.startswith('--'):
                    params.append(f"{k}={v}")
                elif k.startswith('-'):
                    params.append(f"{k} {v}")
                else:
                    params.append(f"-{k} {v}")
        run_cmd(f" ".join(map(str, [gdal_func, *args, *params])), debug=self.debug)

    def _finalize_result(self, out_filepath, label_as, debug=None):
        self.out_filepath = self.data[label_as or 'path'] = out_filepath
        debug = debug or self.debug
        if debug:
            if isinstance(debug, str):
                log_level = debug.lower()
            else:
                log_level = 'debug'
            getattr(logger.a(), log_level)("%s: %s", label_as or get_name(self.out_filepath), self.out_filepath).s()
        if not Path(self.out_filepath).exists():
            raise FileNotFoundError(self.out_filepath)
        return self

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def gdalwarp(self, *srcfiles, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            # if not Path(out_filepath).exists(): os.remove(out_filepath)
            self._run_gdal_cmd('gdalwarp', *srcfiles, out_filepath, **dict(dict(overwrite=True), **kwargs))
        return self._finalize_result(out_filepath, label_as, debug)

    def gdalbuildvrt(self, *gdalfiles, out_filepath, label_as=None, debug=None, **kwargs):
        self._run_gdal_cmd('gdalbuildvrt', out_filepath, *gdalfiles, **dict(dict(overwrite=True), **kwargs))
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_translate(self, src_dataset, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('gdal_translate', src_dataset, out_filepath, **kwargs)
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_merge(self, *input_files, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('gdal_merge.py', *input_files, **dict(o=out_filepath, **kwargs))
        return self._finalize_result(out_filepath, label_as, debug)

    def gdaldem(self, mode, input_dem, out_filepath, color_text_file=None,
                lazy=None, label_as=None, debug=None, **kwargs):
        if mode == 'color-relief': assert color_text_file is not None
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            if color_text_file:
                self._run_gdal_cmd('gdaldem', mode, input_dem, color_text_file, out_filepath, **kwargs)
            else:
                self._run_gdal_cmd('gdaldem', mode, input_dem, out_filepath, **kwargs)
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_calc(self, untagged_expression, out_filepath, equation_info,
                  lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            tags = list(string.ascii_uppercase[:len(equation_info)])
            kwargs = {k.lstrip('-'): v for k, v in kwargs.items()}
            additional_eqn_info = {k: v for k, v in kwargs.items() if k[0].isupper()}

            calc_info = {item_name: dict(tag=tag, path=path) for (item_name, path), tag in
                         zip(dict(equation_info, **additional_eqn_info).items(), tags)}

            tagged_expression = untagged_expression.format(
                **{band_name: v['tag'] for band_name, v in calc_info.items()})
            options = {v['tag']: v['path'] for v in calc_info.values()}

            undashed_options = {'outfile': out_filepath, 'calc': f'"{tagged_expression}"', 'overwrite': True}

            undashed_options.update(kwargs)
            if 'q' in undashed_options:
                q = undashed_options.pop('q')
                undashed_options['quiet'] = q
            for k, v in undashed_options.items():
                options["--" + k] = v
            self._run_gdal_cmd('gdal_calc.py', is_gdal_calc=True, **options)
        return self._finalize_result(out_filepath, label_as, debug)

    def ogr2ogr(self, input_datasource, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('ogr2ogr', out_filepath, input_datasource, **dict(dict(overwrite=True), **kwargs))
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_contour(self, src_filename, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('gdal_contour', src_filename, out_filepath, **kwargs)
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_polygonize(self, raster_file, out_filepath, layer='polygons',
                        connectedness=4, nomask=False, mask=None, band=None, ogr_format=None, fieldname='DN',
                        lazy=None, label_as=None, debug=None, q=None):
        """
        gdal_polygonize.py [-8] [-nomask] [-mask filename] <raster_file> [-b band]
                   [-q] [-f ogr_format] <out_file> [layer] [fieldname]
        :return: self
        """
        args_1 = list()
        if connectedness != 4: args_1.append(f"-{connectedness}")
        if nomask: args_1.append(f"-nomask")
        if mask is not None: args_1.append(f"-mask {mask}")
        args_2 = list()
        if band is not None: args_2.append(f"-b {band}")
        if q is not None: args_2.append("-q")
        if ogr_format is not None: args_2.append(f"-f {ogr_format}")
        args_3 = list()
        if layer: args_3.append(layer)
        if fieldname: args_3.append(fieldname)

        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('gdal_polygonize.py', *args_2, raster_file, *args_2, out_filepath, *args_3)
        return self._finalize_result(out_filepath, label_as, debug)

    def gdal_rasterize(self, src_filename, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
        """
        gdal_rasterize [-b band]* [-i] [-at]
    {[-burn value]* | [-a attribute_name] | [-3d]} [-add]
    [-l layername]* [-where expression] [-sql select_statement]
    [-dialect dialect] [-of format] [-a_srs srs_def] [-to NAME=VALUE]*
    [-co "NAME=VALUE"]* [-a_nodata value] [-init value]*
    [-te xmin ymin xmax ymax] [-tr xres yres] [-tap] [-ts width height]
    [-ot {Byte/Int16/UInt16/UInt32/Int32/Float32/Float64/
            CInt16/CInt32/CFloat32/CFloat64}]
    [-optim {[AUTO]/VECTOR/RASTER}] [-q]
    <src_datasource> <dst_filename>
        """
        lazy = self._get_laziness(lazy, out_filepath)
        if not (lazy and os.path.exists(out_filepath)):
            self._run_gdal_cmd('gdal_rasterize', src_filename, out_filepath, **kwargs)
        return self._finalize_result(out_filepath, label_as, debug)


def get_geo_info(upper_left_xy: Union[list, tuple, np.ndarray],
                 resol_xy: Union[list, tuple, np.ndarray],
                 epsg: int) -> tuple:
    """
    Generates custom geo_info.
    :param upper_left_xy: X and Y coordinates of upper left corner.
    :param resol_xy: Resolution of image in X and Y directions.
    :param epsg: EPSG
    :return: transform and projection (i.e. geo_info).
    """
    transform = np.zeros(6)
    transform[0] = upper_left_xy[0] - resol_xy[0] / 2
    transform[1:3] = resol_xy[0], 0
    transform[3] = upper_left_xy[1] + resol_xy[1] / 2
    transform[4:] = 0, -resol_xy[1]

    spatial_ref = osr.SpatialReference()
    spatial_ref.ImportFromEPSG(epsg)
    projection = spatial_ref.ExportToWkt()
    geo_info = (transform, projection)
    return geo_info



import collections
import functools
import gc
import itertools

import more_itertools as mi
import networkx as nx
import numpy as np
import pandas as pd
from box import Box
from lgblkb_tools import logger
from matplotlib import pyplot as plt, pyplot
from matplotlib.patches import PathPatch
from scipy.spatial.distance import cdist
from shapely import geometry as shg, ops as shops
from sklearn.cluster import KMeans

from lgblkb_tools.geometry.utils.geom_utils import GenericGeometry, get_from_geojson, line_xy, pathify, \
    is_clockwise, generate_cells, min_dist
from .line_utils import TheLine
from .point_utils import ThePoint
from .utils.katana import katana
from .utils.plot_utils import plot_polygon, plot_patches
from .utils.tsp import SimpleTSP


class TheNodeView(object):

    def __init__(self, parent_graph, getter_obj):
        self.parent_graph = parent_graph
        self.getter_obj = getter_obj

    def __getitem__(self, item):
        if type(item) in [list, tuple]:
            return self.getter_obj(*item)
        else:
            return self.getter_obj(item)


class TheGraph(nx.Graph):

    # crop_field.G.nodes[decision_points['inner'][0]]['lines'][1].get_all_points
    def __init__(self, incoming_graph_data=None, **attr):
        super(TheGraph, self).__init__(incoming_graph_data=incoming_graph_data, **attr)
        self.decision_points = collections.OrderedDict()

    def _decision_points_getter(self, decision_key, node_index, *keys):
        result = self.nodes[self.decision_points[decision_key][node_index]]
        for key in keys:
            result = result[key]
        return result

    @property
    def inner_nodes(self):
        return TheNodeView(self, functools.partial(self._decision_points_getter, 'inner'))

    @property
    def border_nodes(self):
        return TheNodeView(self, functools.partial(self._decision_points_getter, 'border'))


class ThePoly(GenericGeometry):

    def __init__(self, polygon, **data):
        super(ThePoly, self).__init__(**data)

        if isinstance(polygon, shg.Polygon):
            self.polygon: shg.Polygon = polygon
        elif isinstance(polygon, ThePoly):
            self.polygon: shg.Polygon = polygon.polygon
        else:
            self.polygon: shg.Polygon = shg.Polygon(polygon)

        if not self.p.is_valid:
            # self.plot()
            # plt.show()
            # raise AssertionError(f"Polygon {self.polygon} is not a valid polygon.")
            # logger.warning(f"Polygon %s is not a valid polygon.",self.polygon)
            pass

        self._xy = None
        self._holes = None
        self._cover_line_length = None
        self._bounds_xy = None
        self._lims = None
        self._is_ok = None
        self._isclockwise = None
        if self.polygon.is_empty:
            self.is_multilinestring = False
        else:
            self.is_multilinestring = isinstance(self.polygon.boundary, shg.MultiLineString)

        self._is_convex = None
        self._convex_hull = None
        self._holes_xy = None
        self._holes = None

        # self.parent=parent
        self._exterior_lines = list()
        self._graph = None
        self._points = list()
        self._all_points = list()
        self._intersection_points = dict()
        self._all_lines = list()
        self._extension_lines = list()
        self._angles_at_vertices = None
        self.data = Box(data)

    @property
    def is_empty(self):
        return self.geometry.is_empty

    def get_simplified(self, tolerance=1e-9):
        return self.__class__(self.geometry.simplify(tolerance))

    @property
    def geometry(self):
        return self.polygon

    # region Class methods:
    @classmethod
    def from_tsp(cls, cities_count=10, seed=None, city_locator=None, **box_kwargs):
        np.random.seed(seed)
        if city_locator is None: city_locator = lambda x: x
        points = city_locator(np.random.rand(cities_count, 2))
        stsp = SimpleTSP(points)
        route = stsp.run(depot=0, show=False)
        the_poly = cls(points[route], **box_kwargs)
        return the_poly

    @classmethod
    def synthesize(cls, cities_count, poly_extent=1000, hole_count=0, seed=None, hole_cities_count=None):
        if seed is None: seed = np.random.randint(0, int(1e6))
        logger.info('seed: %s', seed)
        np.random.seed(seed)
        try_count = 0
        while True:
            try_count += 1
            logger.debug('Try main_poly: %s', try_count)
            seed = np.random.randint(100000000)
            try:
                unholy_field = cls.from_tsp(cities_count=cities_count, seed=seed,
                                            city_locator=lambda ps: ps * poly_extent)
                if hole_count < 1: return unholy_field
                # holes_current_count=0
                holes = list()

                for i in range(1000):
                    # .plot()
                    seed = np.random.randint(100000000)
                    hole = cls.from_tsp(cities_count=hole_cities_count or np.random.randint(4, 10), seed=seed,
                                        city_locator=lambda ps: ps * np.random.randint(poly_extent * 0.2,
                                                                                       poly_extent * 0.5) +
                                                                np.random.randint(poly_extent * 0.1, poly_extent * 0.2,
                                                                                  2))
                    # holy_poly=shg.Polygon(unholy_field.xy,[hole.xy,*map(lambda h:h.xy,holes)])
                    # hole.plot()

                    # logger.debug('unholy_field.p.contains(hole.p): %s',unholy_field.p.contains(hole.p))
                    # logger.debug('unholy_field.p.intersects(hole.p): %s',unholy_field.p.intersects(hole.p))
                    if not unholy_field.p.contains(hole.p): continue

                    # 	logger.debug('Contains!!!')
                    # if unholy_field.p.intersects(hole.p):
                    # 	logger.debug('Intersects!!!')
                    # if (not unholy_field.p.contains(hole.p)) or unholy_field.p.intersects(hole.p):
                    # 	continue
                    # hole.plot()
                    # unholy_field.plot()
                    # plt.show()
                    hole_friendly = True
                    for other_hole in holes:
                        # logger.debug('other_hole.p.intersects(hole.p):\n%s',other_hole.p.intersects(hole.p))
                        # logger.debug('other_hole.p.within(hole.p):\n%s',other_hole.p.within(hole.p))
                        if other_hole.p.intersects(hole.p) or other_hole.p.within(hole.p):
                            hole_friendly = False
                            break
                    if not hole_friendly: continue

                    holes.append(hole)
                    holy_poly = shg.Polygon(unholy_field.xy, list(map(lambda x: x.xy, holes)))  # .buffer(0)

                    # if not holy_poly.is_valid:
                    # 	holes.pop()
                    # 	continue

                    holy_field = cls(holy_poly, area=holy_poly.area)

                    logger.debug('Proper hole created at i = %s', i)
                    # holy_field.plot()
                    # unholy_field.plot()
                    # plt.show()

                    if len(holy_field.holes) == hole_count:
                        # if holes_current_count==hole_count:
                        gc.collect()
                        return holy_field
                gc.collect()
            # if holy_field.is_multilinestring:
            # 	# holes_current_count+=1
            # 	# logger.debug('holy_field:\n%s',holy_field)
            # 	# the_hole.plot()
            # 	# unholy_field.plot()
            # 	# plt.show()
            # 	# holes.append(the_hole)
            # 	# logger.debug('Hole %s added.',hole_count)
            # 	# holy_field=cls(shg.Polygon(unholy_field.xy,holes=list(map(lambda x:x.xy,holes))))#.plot(c='r')
            # 	if len(holy_field.holes)==hole_count:
            # 		# if holes_current_count==hole_count:
            # 		return holy_field
            except Exception as exc:
                logger.debug('exc:\n%s', exc)
                raise
                pass

    @classmethod
    def from_geojson(cls, filepath, **box_kwargs):
        return cls(get_from_geojson(filepath)[0], **box_kwargs)

    @classmethod
    def as_valid(cls, raw_polygon):
        raw_poly = ThePoly(raw_polygon)
        # for i, valid_point in enumerate(raw_poly.points):
        # 	valid_point.plot(f'P-{i}')
        invalid_point_indexes = list()
        for i, angle in enumerate(raw_poly.angles_at_vertices):
            # logger.debug('angle: %s',angle)
            if (180 - angle) % 180 < min_dist:
                # logger.info('i: %s',i)
                invalid_point_indexes.append(i)
        valid_points = [x for i, x in enumerate(raw_poly.points) if not i in invalid_point_indexes]
        valid_poly = cls(valid_points)
        for hole in raw_poly.holes:

            valid_hole_poly = cls.as_valid(hole)
            try:
                diff_result = valid_poly.geometry.difference(valid_hole_poly.geometry)
            # if isinstance(diff_result,shg.MultiPolygon):
            # 	for diff_result_i in diff_result:
            # 		ThePoly(diff_result_i).plot()
            # 	plt.show()

            except:
                # # raw_poly.plot()
                # valid_poly.plot()
                # valid_hole_poly.plot('hole')
                # plt.show()

                # raise
                diff_result = valid_poly.geometry.buffer(1e-9).difference(valid_hole_poly.geometry)

            try:
                valid_poly = cls(diff_result)
            except:
                logger.warning('diff_result:\n%s', diff_result)
                valid_poly = cls(diff_result.buffer(1e-9))
            # raise

        # invalid_line_indexes.append(i-1)
        # for neighbor_index in [i-1,i+1]:
        # # neighbor_index=i-1
        # 	if raw_poly[i].angle-raw_poly[neighbor_index].angle<min_dist:
        # 		invalid_line_indexes.append(neighbor_index)
        # 		break

        # for i, valid_point in enumerate(valid_points):
        # 	valid_point.plot(f'P-{i}')

        # .plot(c='r')
        if len(valid_poly) != len(raw_poly):
            return cls.as_valid(valid_poly)
        else:
            return valid_poly

    # endregion

    def original_poly(self):
        if not self.data.get('__toy_poly_scale__'):
            raise AttributeError('The polygon is not a toy. Original polygon does not exist.', dict(polygon=str(self)))
        return self.__class__(self.xy / self.data['__toy_poly_scale__'] + self.data['__toy_poly_mean_xy__'])

    def toy_poly(self, scale=1e-3):
        mean_xy = np.mean(self.xy, axis=0)
        zero_mean_xy = self.xy - mean_xy
        # print(zero_mean_xy)
        scaled_xy = zero_mean_xy * scale
        crop_field = self.__class__(scaled_xy, __toy_poly_scale__=scale, __toy_poly_mean_xy__=mean_xy)  # .plot()
        return crop_field

    @property
    def p(self):
        return self.polygon

    def plot(self, text=None, hole_opts=None, **kwargs):
        # gmtr.plot_polygon(self.polygon,**dict(dict(c='gray'),**kwargs))
        plot_polygon(self.polygon, **kwargs)
        plot_patches(map(lambda h: h.p, self.holes), **dict(dict(alpha=0.7), **(hole_opts or {})))
        if text is not None: plt.text(*np.array(self.polygon.centroid.xy), s=text)
        return self

    def buffer(self, distance, resolution=16, quadsegs=None, cap_style=shg.CAP_STYLE.round,
               join_style=shg.JOIN_STYLE.round, mitre_limit=5.0):
        return self.__class__(self.polygon.buffer(distance, resolution=resolution, quadsegs=quadsegs,
                                                  cap_style=cap_style, join_style=join_style, mitre_limit=mitre_limit))

    @property
    def reversed(self):
        return self.__class__(self.xy[::-1], **self.data)

    @property
    def holes(self):
        if self._holes is None:
            if isinstance(self.p.boundary, shg.MultiLineString):
                self._holes = [self.__class__(x, parent=self) for x in self.p.boundary[1:]]
            else:
                self._holes = []
        return self._holes

    @property
    def holes_xy(self):
        if self._holes_xy is None:
            if isinstance(self.polygon.boundary, shg.MultiLineString):
                self._holes_xy = [line_xy(x) for x in self.polygon.boundary[1:]]
            else:
                self._holes_xy = []
        return self._holes_xy

    @property
    def all_xy(self):
        return np.concatenate([self.xy[:-1], *[x[:-1] for x in self.holes_xy]])

    @property
    def xy(self):
        if self._xy is None:
            if self.is_multilinestring:
                self._xy = line_xy(self.polygon.boundary[0])
            else:
                self._xy = line_xy(self.polygon.boundary)
        return self._xy

    @property
    def bounds_xy(self):
        if self._bounds_xy is None:
            self._bounds_xy = line_xy(self.polygon.envelope.boundary)
        return self._bounds_xy

    @property
    def exterior_lines(self):
        if not self._exterior_lines: self._exterior_lines = self.get_exterior_lines()
        return self._exterior_lines

    @property
    def points(self):
        if not self._points:
            self._points = [ThePoint(xy) for xy in self.xy[:-1]]
        return self._points

    @property
    def all_points(self):
        if not self._all_points:
            self._all_points = [ThePoint(xy) for xy in self.all_xy]
        return self._all_points

    @property
    def all_lines(self):
        """

        :rtype: [TheLine]
        """
        if not self._all_lines:
            self._all_lines = [*self.exterior_lines, *mi.flatten([x.exterior_lines for x in self.holes])]
        return self._all_lines

    def get_exterior_lines(self, as_the_line=True, show=0):
        lines = list()
        if as_the_line:
            line_getter = lambda _line: TheLine(_line)
        else:
            line_getter = lambda _line: _line

        if self.is_multilinestring:
            is_ok = False
            for _ in self.polygon.boundary:
                is_ok = True
                break
            if not is_ok:
                logger.critical('self.geometry:\n%s', self.geometry)
                for p in self.points: p.plot()
                logger.critical('len(self.points): %s', len(self.points))
                plt.show()

        current_chunk = self.polygon.boundary[0] if self.is_multilinestring else self.polygon.boundary
        for xy in self.xy[1:-1]:
            # ThePoint(xy).plot()
            line, current_chunk = shops.split(current_chunk, shg.Point(xy))
            lines.append(line_getter(line))
        lines.append(line_getter(current_chunk))
        if show:
            for i, line in enumerate(lines):
                if as_the_line:
                    line.plot()
                else:
                    TheLine(line).plot()

        # TheLine(line=line).plot(**plot_kwargs)
        # logger.debug('line.xy: %s',line.xy)
        # plt.plot(*line.xy,**plot_kwargs)
        # plt.plot(*line.parallel_offset(5).xy)
        return lines

    @property
    def cover_line_length(self):
        if self._cover_line_length is None:
            # xy=self.__class__(self.polygon.envelope).xy
            delta = self.bounds_xy[2] - self.bounds_xy[0]
            self._cover_line_length = np.linalg.norm(delta) * 1.01
        return self._cover_line_length

    # def get_field_lines(self,offset_distance,border_index=None,show=0):
    # 	# if border_index is None: slicer=lambda x:x[:]
    # 	# elif isinstance(border_index,Iterable): slicer=lambda x:[x[_i] for _i in border_index]
    # 	# else: slicer=lambda x:x[border_index]
    # 	if border_index!=0 and not border_index: checker=lambda x:True
    # 	elif isinstance(border_index,Iterable): checker=lambda x:x in border_index
    # 	else: checker=lambda x:x==border_index
    # 	borderlines=[TheLine(x,width=offset_distance/2) for x in
    # 	             self.buffer(-1e-8,join_style=shg.JOIN_STYLE.mitre).get_exterior_lines(as_the_line=False,show=0)]
    # 	field_lines_data=collections.defaultdict(list)
    # 	for i,borderline in enumerate(borderlines):
    # 		if not checker(i): continue
    # 		counter=0
    # 		offset_lines=borderline.get_field_lines(self)
    # 		offset_lines.extend(borderline.offset_by(offset_distance,int(np.ceil(self.cover_line_length/offset_distance))))
    # 		offset_lines.extend(borderline.offset_by(-offset_distance,int(np.ceil(self.cover_line_length/offset_distance))))
    #
    # 		for offset_line in offset_lines:
    # 			# offset_line.plot()
    #
    # 			field_lines=offset_line.get_field_lines(self)
    # 			if not field_lines:
    # 				continue
    # 			for field_line in field_lines:
    # 				if not field_line: continue
    #
    # 				counter+=1
    # 				field_lines_data[i].append(field_line)
    # 	# plt.show()
    # 	# if i==0 or i==1:
    # 	# 	field_line.plot('100',c='magenta',alpha=0.5)
    # 	# logger.info('counter: %s',counter)
    # 	fl_counters=[len(x) for x in field_lines_data.values()]
    # 	logger.debug('# of field lines along borders: %s',fl_counters)
    #
    # 	if show:
    # 		# fig=plt.gcf()
    # 		# fig.set_size_inches(16,9,forward=True)
    # 		for i,b in enumerate(borderlines):
    # 			if not checker(i): continue
    # 			# self.plot()
    # 			boundary_label=f'Boundary_{i}'
    # 			b.plot('1001',text=boundary_label)
    # 			# df=pd.DataFrame(field_lines_data[i],columns=[boundary_label])
    # 			# df['length']=df.applymap(lambda x:x.line.length)
    # 			# df[boundary_label].map(lambda x:x.plot('100',c='magenta',alpha=0.5))
    # 			for field_line in field_lines_data[i]:
    # 				field_line.plot()
    #
    # 	# df['length'].plot.hist()
    # 	# plt.text(*np.array(self.polygon.centroid.xy),df.shape[0])
    # 	# plt.show()
    # 	return field_lines_data

    @property
    def lims(self):
        if self._lims is None:
            self._lims = np.array([self.bounds_xy[0], self.bounds_xy[2]])
        return self._lims

    # region Discretization methods:
    def pixelize(self, savename='', show=0, **fig_kwargs):
        plt.clf()
        fig = pyplot.figure(**dict(dict(num=1, figsize=(4, 4), dpi=50, frameon=False), **fig_kwargs))
        ax = plt.Axes(fig, [0., 0., 1., 1.], frameon=False)
        ax.set_axis_off()
        fig.add_axes(ax)
        # ax=fig.add_subplot(111)

        path = pathify(self.polygon)
        # patch=PathPatch(path,facecolor='#cccccc',edgecolor='#999999')
        patch = PathPatch(path, facecolor='k', edgecolor='k')

        ax.add_patch(patch)
        lims = np.array([self.bounds_xy[0], self.bounds_xy[2]])
        ax.set_xlim(*lims[:, 0])
        ax.set_ylim(*lims[:, 1])
        ax.set_aspect('1.0')
        fig.canvas.draw()
        X = np.array(fig.canvas.renderer._renderer)
        if savename: fig.savefig(savename)
        plt.close(fig)
        if show:
            fig2 = plt.figure()
            ax2 = fig2.add_subplot(111, frameon=False)
            ax2.imshow(X)
            plt.show()
        return X

    @logger.trace()
    def katanize(self, threshold, show=0, **kwargs):
        polygons = katana(self.polygon, threshold)
        if show:
            for p in polygons: plot_polygon(p, **kwargs)
        return polygons

    # def get_interior_points(self,distance,border_index=None,show=0):
    # 	lines=collections.defaultdict(list)
    # 	if border_index is None:
    # 		# field_box=ThePoly(self.polygon.envelope)
    # 		for i,box_lines in self.get_field_lines(distance,[0,1]).items():
    # 			for bl in box_lines:
    # 				lines[i].append(bl)
    # 	else:
    # 		for i,field_lines in self.get_field_lines(distance,border_index).items():
    # 			for bl in field_lines:
    # 				lines[i].append(bl)
    # 	the_data=pd.DataFrame(lines[(border_index[-1] if border_index is not None else 1)],columns=['line2s'])
    # 	intersection_points=list()
    # 	for l1 in lines[0]:
    # 		points=the_data['line2s'].apply(lambda l2:l2.line.intersection(l1.line))
    # 		not_empty_ones=points.apply(lambda x:isinstance(x,shg.Point))
    # 		good_points=points[not_empty_ones]
    # 		if good_points.empty: continue
    # 		intersection_points.extend(good_points.tolist())
    # 	ps=pd.Series(intersection_points)
    # 	out=np.array(list(mi.flatten(ps.apply(lambda p:line_xy(p)))))
    # 	if show:
    # 		plt.scatter(out[:,0],out[:,1])
    # 		pass
    # 	return out

    @logger.trace()
    def generate_clusters(self, grid_resolution, n_clusters, show='', random_state=1):
        # interior_points=self.get_interior_points(distance=grid_resolution)
        interior_points = self.generate_grid_points(grid_resolution=grid_resolution)
        logger.debug('len(interior_points): %s', len(interior_points))
        n_clusters = min(n_clusters, len(interior_points))
        logger.debug('n_clusters: %s', n_clusters)
        kmeans = KMeans(n_clusters=n_clusters, random_state=random_state)
        kmeans.fit(interior_points)
        final_centers = kmeans.cluster_centers_.copy()
        for i, center in enumerate(kmeans.cluster_centers_):
            if not ThePoint(center).geometry.within(self.geometry):
                grid_points = interior_points[kmeans.labels_ == i]
                distances = cdist(kmeans.cluster_centers_[i].reshape(-1, 2), grid_points, )
                closest_grid_point = grid_points[np.argmin(distances)]
                final_centers[i] = closest_grid_point

        if show:
            plot_clusters, plot_self = 1, 0
            if len(show) == 1:
                plot_clusters = int(show)
            elif len(show) == 2:
                plot_clusters, plot_self = [int(x) for x in show]
            if plot_self: self.plot()
            if plot_clusters:
                plt.scatter(kmeans.cluster_centers_[:, 0], kmeans.cluster_centers_[:, 1], c='k')
        return final_centers

    @logger.trace()
    def generate_grid_points(self, grid_resolution):
        corners = self.bounds_xy[[0, 2]]
        x_grid_info = np.arange(*corners[:, 0], step=grid_resolution)
        y_grid_info = np.arange(*corners[:, 1], step=grid_resolution)
        grid_x, grid_y = np.meshgrid(x_grid_info, y_grid_info)
        xy_coors = np.stack([grid_x.ravel(), grid_y.ravel()]).T
        # print(xy_coors)
        xy_df = pd.DataFrame(xy_coors, columns=['x', 'y'])
        xy_df['inside'] = xy_df.apply(lambda coor: self.polygon.contains(shg.Point([coor.x, coor.y])), axis=1)
        xy_df = xy_df[xy_df['inside']].drop(columns=['inside'])
        logger.debug('xy_df.shape: %s', xy_df.shape)
        return xy_df.to_numpy()

    def to_numpy(self, resolution):
        logger.info('self.lims: %s', self.lims)

        delta = self.lims[1] - self.lims[0]
        logger.info('delta: %s', delta)
        pixel_count = np.ceil(delta / resolution)[::-1].astype(int)
        logger.info('pixel_count: %s', pixel_count)
        img_array = np.zeros(shape=pixel_count)
        interior_points = np.round(self.generate_grid_points(resolution), decimals=1)[:, [1, 0]]
        positions_in_grid = np.floor(interior_points / resolution).astype(int)
        positions_in_grid = np.clip(positions_in_grid, a_min=[0, 0], a_max=np.array(img_array.shape) - 1)
        img_array[positions_in_grid[:, 0], positions_in_grid[:, 1]] = 1
        return img_array

    # endregion

    # region Booleans:
    @property
    def is_clockwise(self):
        if self._isclockwise is not None: return self._isclockwise
        x = self.xy[:, 0]
        y = self.xy[:, 1]
        diff_x = x[1:] - x[:-1]
        sum_y = y[:-1] + y[1:]
        area = np.sum(diff_x * sum_y)
        # logger.info('x:\n%s',x)
        # logger.info('diff_x:\n%s',diff_x)
        # logger.info('y:\n%s',y)
        # logger.info('sum_y:\n%s',sum_y)
        # logger.info('area: %s',area)
        if area > 0:
            self._isclockwise = True
        elif area < 0:
            self._isclockwise = False
        else:
            logger.warning('self.polygon:\n%s', self.polygon)
            # self._isclockwise=None
            self.plot()
            plt.show()
            raise NotImplementedError('Area is zero.')
        return self._isclockwise

    @property
    def is_ok(self):
        if self._is_ok is not None: return self._is_ok
        if not self.is_clockwise:
            logger.info('self.polygon:\n%s', self.polygon)
            # logger.info('"self" is clockwise, but should be counter-clockwise')
            logger.info('"self" is counter-clockwise, but should be clockwise')
            self._is_ok = False
            return self._is_ok
        for interior in self.polygon.interiors:
            interior_poly = shg.Polygon(interior)
            if is_clockwise(interior_poly):
                logger.info('Interior polygon:\n%s', interior_poly)
                logger.info('"Interior polygon" is clockwise, but should be counter-clockwise')
                # logger.info('"Interior polygon" is counter-clockwise, but should be clockwise')
                self._is_ok = False
                return self._is_ok
        self._is_ok = True
        return self._is_ok

    def touches(self, xy):
        return self.polygon.touches(shg.Point(xy))

    def as_noncollinear(self):
        non_collinear_points = list()
        for i_vertex, angle_at_vertex in enumerate(self.angles_at_vertices):
            if abs(angle_at_vertex - 180) < min_dist: continue
            non_collinear_points.append(self.points[i_vertex])
        non_collinear_field = self.__class__(non_collinear_points)
        return non_collinear_field

    @property
    def is_convex(self):
        if self._is_convex is None:
            self._is_convex = True
            for angle_at_vertex in self.angles_at_vertices:
                if angle_at_vertex > 180 + min_dist:
                    self._is_convex = False
                    break
        # if abs(self.p.area-self.convex_hull.p.area)<min_dist:
        # 	self._is_convex=True
        # else:
        # 	self._is_convex=False
        return self._is_convex

    # endregion

    # region Boustrophedon algorithm:
    def get_connectivity_lines_at(self, xy, **the_line_kwargs):
        line = TheLine([xy, xy + [0, 1]])
        cover_line = line.get_cover_line(self.cover_line_length)
        res = cover_line.line.intersection(self.polygon)
        outs = list()
        if isinstance(res, shg.GeometryCollection):
            if isinstance(res[0], shg.Point): return []
            for geom in res:
                if isinstance(geom, shg.LineString) and geom.touches(shg.Point(*xy)):
                    outs.append(TheLine(geom, **the_line_kwargs))
        elif isinstance(res, shg.MultiLineString):
            for geom in res:
                if geom.touches(shg.Point(*xy)):
                    outs.append(TheLine(geom, **the_line_kwargs))
        return outs

    def get_boustrophedon_lines(self, **the_line_kwargs):
        outs = list()
        for xy in self.xy:
            lines = self.get_connectivity_lines_at(xy, **the_line_kwargs)
            if lines: outs.append(lines)
        for holexy in self.holes_xy:
            xy_sorted = sorted(holexy[:-1], key=lambda x: x[0])
            for xy in [xy_sorted[0], xy_sorted[-1]]:
                lines = self.get_connectivity_lines_at(xy, **the_line_kwargs)
                if lines: outs.append(lines)
        return outs

    def get_boustrophedon_cells(self, as_the_poly=True, show=0, **plot_kwargs):
        assert self.is_ok
        line_groups = self.get_boustrophedon_lines()
        merged_lines = list()
        for lines in line_groups:
            line = shops.linemerge(shg.MultiLineString([x.line for x in lines]))
            merged_lines.append(line)
        merged_lines.sort(key=lambda x: x.xy[0])
        cells = generate_cells(self.polygon, merged_lines)
        logger.info('len(cells): %s', len(cells))
        if show:
            for i, c in enumerate(cells):
                self.__class__(c.buffer(-0.5)).plot(**dict(dict(c=None, text=str(i)), **plot_kwargs))
        if as_the_poly: cells = [self.__class__(x) for x in cells]
        return cells

    def get_connectivity_graph(self, as_the_poly=True, show=''):
        G = nx.Graph()
        cells = self.get_boustrophedon_cells(as_the_poly=as_the_poly, show=int(show))

        G.add_nodes_from(cells)

        if as_the_poly:
            polymaker = lambda x: x.polygon
        else:
            polymaker = lambda x: x
        connections = list()
        bufferer = lambda x: x.buffer(1e-3, cap_style=shg.CAP_STYLE.square)

        for c1, c2 in itertools.combinations(cells, 2):
            c1_poly: shg.Polygon = polymaker(c1)
            c2_poly: shg.Polygon = polymaker(c2)

            merged_mpoly = bufferer(shg.MultiPolygon([c1_poly, c2_poly]))
            if isinstance(merged_mpoly, shg.Polygon):
                intersection_area = bufferer(c1_poly).intersection(bufferer(c2_poly)).area
                if intersection_area < 1e-5: continue
                connections.append([c1, c2])
                G.add_edge(c1, c2)
                if show:
                    TheLine(shg.LineString([c1_poly.centroid, c2_poly.centroid])).plot(c='gray', ls='--')
        return cells, G

    # endregion

    # def get_convex_hull(self):
    # 	if self._convex_hull is None:
    # 		hull=spatial.ConvexHull(self.xy)
    # 		vertices=np.array(sorted(hull.vertices))
    # 		# convex_poly=ThePoly(hull.points[hull.vertices]).plot(c='b',alpha=0.5)
    # 		convex_poly=ThePoly(self.xy[vertices])  #.plot(c='b',alpha=0.5)
    # 		self._convex_hull=convex_poly
    # 	return self._convex_hull
    @property
    def convex_hull(self):
        if self._convex_hull is None:
            # logger.debug('self.xy:\n%s',self.xy)
            if self.p.convex_hull.is_empty:
                # self.plot()
                # plt.show()
                logger.warning('Empty hull')
                self._convex_hull = ThePoly(shg.Polygon())
            else:
                self._convex_hull = ThePoly(self.p.convex_hull, **self.data)
        return self._convex_hull

    def __contains__(self, item):
        if isinstance(item, ThePoint):
            return item in self.points
        elif isinstance(item, ThePoly):
            return self.geometry.contains(item.geometry) or self.geometry.intersects(
                item.geometry) or self.geometry.touches(item.geometry)

    def __eq__(self, other):
        # if isinstance(other,ThePoly):
        # 	return self.id==other.id
        # else:
        # 	return self.id==ThePoly(other).id
        # is_equal=True
        if isinstance(other, ThePoly):
            other_points = other.points
        else:
            other_points = ThePoly(other).points

        other_points_copy = other_points[:]
        for point in self.points:
            if not point in other_points_copy:
                return False
            other_points_copy.remove(point)
        if other_points_copy: return False
        return True

    def __len__(self):
        return len(self.xy) - 1

    def _add_node(self, G, xy, parent_lines, is_intermediate=False):
        node = ThePoint(xy)
        l1, l2 = parent_lines
        if is_intermediate:
            if not node in l1:
                l1.intermediate_points[node.associate_with_lines(l1, l2)] = l2
            if not node in l2:
                l2.intermediate_points[node.associate_with_lines(l1, l2)] = l1
        G.add_node(node, lines=parent_lines)
        return node

    def _add_edge(self, G, line):
        G.add_edge(line[0], line[1], line=line)
        return line

    # @logger.trace()
    def generate_graph(self, parent_graph=None):
        G = parent_graph or TheGraph()
        # vertex_angles=self.get_angles_at_vertices()
        # if self.parent: vertex_angles=360-vertex_angles

        # point_kwargs=dict()
        # point_kwargs.update(box_kwargs)
        # point_kwargs.update(point_opts or {})

        for i in range(len(self.exterior_lines)):
            # next_index=(i+1) if i+1==len(lines) else 0
            # node=ThePoint(self.xy[i])
            # G.add_node(node,lines=[lines[i],lines[i-1]])
            self._add_node(G, self.xy[i], parent_lines=[self.exterior_lines[i], self.exterior_lines[i - 1]])
            # next_node=ThePoint(self.xy[next_index])
            # G.add_edge(node,next_node,line=lines[i])
            pass
            # self._add_edge(G,self.exterior_lines[i])
            pass

        for hole in self.holes:
            hole.generate_graph(G)

        # for i, xy in enumerate(self.xy):
        # 	G.add_node(i,)

        # for i,line in enumerate(lines):
        # 	# G.add_edge(*map(ThePoint,line.xy),line=line)
        # 	if i+1==len(lines):
        # 		G.add_edge(i,0,line=line)
        # 	else:
        # 		G.add_edge(i,i+1,line=line)
        return G

    @property
    def G(self):
        if self._graph is None: self._graph = self.generate_graph()
        return self._graph

    def generate_intersection_points(self, lines):
        intersection_points = list()
        for l1, l2 in itertools.combinations(lines, 2):
            if l1[0] == l2[0]: continue
            result = l1.line.intersection(l2.line)
            if result.is_empty: continue
            the_point = ThePoint(result)
            self._add_node(self.G, the_point, [l1, l2], is_intermediate=True)
            # self.G.add_node(the_point,lines=[l1,l2])
            intersection_points.append(the_point)
        # intersection_points=set(intersection_points)
        logger.info('intersection_points count: %s', len(intersection_points))
        return intersection_points

    def get_angles_at_vertices(self):
        vectors = self.xy[1:] - self.xy[:-1]
        line_angles = np.arctan2(vectors[:, 1], vectors[:, 0]) * 180 / np.pi
        line_angles = np.append(line_angles, line_angles[0])
        angle_diff = ((line_angles[1:] - line_angles[:-1]) + 180) % 360
        angle_at_vertices = np.zeros(angle_diff.shape)
        angle_at_vertices[0] = angle_diff[-1]
        angle_at_vertices[1:] = angle_diff[:-1]
        if self.is_clockwise:
            return angle_at_vertices
        else:
            return 360 - angle_at_vertices

    @property
    def angles_at_vertices(self):
        if self._angles_at_vertices is None:
            self._angles_at_vertices = self.get_angles_at_vertices()
        return self._angles_at_vertices

    def __getitem__(self, item):
        # logger.debug('item: %s',item)
        if type(item) in (np.ndarray, tuple, list):
            # if isinstance(item,np.ndarray): item=item.ravel()
            u, v = item
            return self.G.nodes[u, v]
        elif isinstance(item, slice):
            return [self[ii] for ii in range(*item.indices(len(self)))]
        elif isinstance(item, int):
            return self.exterior_lines[item]
        else:
            return super().__getitem__(item)

    def __hash__(self):
        return hash(self.geometry.wkt)


class TheChunk(ThePoly):

    def __init__(self, polygon, parent, **data):
        super(TheChunk, self).__init__(polygon=polygon, **data)
        self.parent: ThePoly = parent
        self.touches_parent = self.parent.p.exterior.touches(self.p)  # if parent else False
        self.baseline = None
        if self.touches_parent:
            parent_lines = self.parent.get_exterior_lines()
            for chunk_line in self.get_exterior_lines():
                if chunk_line not in parent_lines:
                    self.baseline = chunk_line
                    # chunk_line.plot(text='baseline')
                    break

            assert self.baseline is not None



import datetime
import logging
import sys
from logging.handlers import RotatingFileHandler
from timeit import default_timer as timer

import wrapt
# class TheLogger(object):
# 	def __init__(self,logger_name):
# 		pass
#
# def get_console_handler():
# 	console_handler=logging.StreamHandler(sys.stdout)
# 	console_handler.setFormatter(FORMATTER)
# 	return console_handler
#
# def get_file_handler():
# 	file_handler=TimedRotatingFileHandler(LOG_FILE,when='midnight',backupCount=3)
# 	file_handler.setFormatter(FORMATTER)
# 	return file_handler
#
# def get_logger(logger_name):
# 	logger=logging.getLogger(logger_name)
# 	logger.setLevel(logging.DEBUG)  # better to have too much log than not enough
# 	logger.addHandler(get_console_handler())
# 	# logger.addHandler(get_file_handler())
# 	# with this pattern, it's rarely necessary to propagate the error up to parent
# 	logger.propagate=False
# 	return logger
from box import Box
from colorlog import ColoredFormatter
from python_log_indenter import IndentedLoggerAdapter

# FORMATTER=logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
# LOG_FILE="lgblkb_tools.log"

colored_log_format = f"%(log_color)s%(asctime)s %(levelname)5.5s %(message)s"
# colored_log_format=f"%(asctime)s — %(name)s — %(levelname)s — %(message)s"
base_log_indent_settings = dict(indent_char='| ', spaces=1)
# date_fmt="%m-%d %H:%M:%S"
date_fmt = "%H:%M:%S"
default_log_format = colored_log_format.replace(r'%(log_color)s', '')

# region log_func_mapper:
log_func_mapper = Box()
log_func_mapper[logging.DEBUG] = lambda _logger: _logger.debug
log_func_mapper[logging.INFO] = lambda _logger: _logger.info
log_func_mapper[logging.WARNING] = lambda _logger: _logger.warning
log_func_mapper[logging.CRITICAL] = lambda _logger: _logger.critical
log_func_mapper[logging.ERROR] = lambda _logger: _logger.error
# endregion

log_sayer = lambda _logger, log_level: log_func_mapper[log_level](_logger)


def generate_handle():
    @wrapt.decorator
    def _handle_generator(wrapped, instance, args, kwargs):
        log_detailer = lambda *_args, level=logging.DEBUG, log_format='', **_kwargs: (level, log_format, _args, _kwargs)
        the_level, the_format, args, kwargs = log_detailer(*args, **kwargs)
        instance.create_handler(wrapped(*args, **kwargs), level=the_level, log_format=the_format)
        instance.info('Log handle %s created at %s', instance.logger.handlers[-1].__class__.__name__,
                      datetime.datetime.now())
        return instance
    
    return _handle_generator


# noinspection PyUnusedLocal
class TheLogger(IndentedLoggerAdapter):
    def __init__(self, logger_name, level=logging.DEBUG, extra=None, auto_add=True, propagate=False, _logger=None,
                 **kwargs):
        if _logger is None:
            _logger = logging.getLogger(logger_name)
        _logger.setLevel(level)
        super(TheLogger, self).__init__(_logger, extra=extra, auto_add=auto_add,
                                        **dict(base_log_indent_settings, **kwargs))
        _logger.propagate = propagate
    
    def create_handler(self, handler: logging.Handler, level=logging.DEBUG, log_format=''):
        if log_format:
            if type(handler) is logging.StreamHandler:
                log_formatter_class = ColoredFormatter
            else:
                log_formatter_class = logging.Formatter
            log_formatter = log_formatter_class(log_format, date_fmt)
        else:
            # if type(handler) is logging.StreamHandler:
            # 	log_formatter=ColoredFormatter(colored_log_format,date_fmt)
            # else:
            log_formatter = logging.Formatter(default_log_format, date_fmt)
        handler.setFormatter(log_formatter)
        handler.setLevel(level=level)
        self.logger.addHandler(handler)
        return self
    
    @generate_handle()
    def to_rotate(self, log_filepath, maxBytes=2.5e6, backupCount=14, **kwargs):
        return RotatingFileHandler(filename=log_filepath, maxBytes=int(maxBytes), backupCount=backupCount)
    
    def trace(self, level=logging.DEBUG, skimpy: bool = False, verbose=False):
        log_say = log_sayer(self, level)
        
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            if skimpy:
                log_say('Performing "%s"...', wrapped.__name__)
            else:
                log_say('Running "%s":', wrapped.__name__)
            if verbose:
                logger.debug('args_count: %s', len(args)).a(3)
                for arg in args:
                    logger.debug('arg: %s', arg)
                logger.s(3)
                logger.debug('kwargs_count: %s', len(kwargs)).a(3)
                for k, v in kwargs.items():
                    logger.debug('%s: %s', k, v)
                logger.s(3)
            start = timer()
            self.add()
            out = wrapped(*args, **kwargs)
            if not skimpy: log_say('Done "%s". Duration: %.3f sec.', wrapped.__name__, timer() - start)
            self.sub()
            return out
        
        return wrapper
    
    @generate_handle()
    def to_file(self, log_filepath, **kwargs):
        return logging.FileHandler(log_filepath, **kwargs)
    
    @generate_handle()
    def to_stream(self, stream=None):
        return logging.StreamHandler(stream or sys.stdout)


logger: TheLogger = TheLogger('lgblkb_logger').to_stream()
logger.info('Log start datetime: %s', datetime.datetime.now())
